#!/usr/bin/env python

from math import exp, pi, sqrt
from dolfin import *
from dolfin_adjoint import *
import libadjoint

parameters["std_out_all_processes"] = False
dolfin.parameters["adjoint"]["record_all"] = True
dolfin.parameters["adjoint"]["fussy_replay"] = True

mesh = Mesh("basin.xml"); mesh.order(); mesh.init()

V = VectorFunctionSpace(mesh, "DG", 1)
P = FunctionSpace(mesh, "CG", 2)
Z = MixedFunctionSpace([V, P])

# The only boundary conditions are free-slip on velocity.

r0 = 250000
H = 5
g = 10
f = 1.0313e-04

c = sqrt(g*H)
LR = c / f

T = 4*2*pi*r0/c
dt = T/4000
T = dt*1

theta = 0.5

u_pvd = File("results/velocity.pvd")
p_pvd = File("results/layer_thickness.pvd")

def compute_ic():
  class IC(Expression):
    def eval(self, values, X):
      r = (X[0]**2+X[1]**2)**0.5

      if r > 0.0001:
        values[0] = -0.05*c*exp((r-r0)/LR)*X[0]/r*X[1]/r
        values[1] =  0.05*c*exp((r-r0)/LR)*X[0]/r*X[0]/r
        values[2] =  0.05*exp((r-r0)/LR)*X[0]/r
      else:
        values[0] = 0.0
        values[1] = 0.0
        values[2] = 0.0

    def value_shape(self):
      return (3,)

  z = Function(Z, name="State")
  z.assign(interpolate(IC(), Z))

  return z

def store(z, t):
  t = float(t)
  if MPI.process_number() == 0:
    print "Storing variables at t=%s" % t

  (u, p) = z.split()

  u_pvd << (u, t)
  p_pvd << (p, t)

def Dt(z_old, z_new, dt):
  return (z_new - z_old)/Constant(dt)

def cn(z_old, z_new): # Crank-Nicolson
  return (1 - theta)*z_old + theta*z_new

def coriolis(u):
  return f * as_vector([-u[1], u[0]])

def main(ic, annotate=False):
  t = 0

  if annotate:
    store(ic, t=t)

  z_old = ic
  (u_old, p_old) = split(z_old)

  z_new = TrialFunction(Z)
  (u_new, p_new) = split(z_new)

  z = Function(ic, name="NewState")
  z.assign(z_old, annotate=annotate)
  (u, p) = split(z)

  (u_test, p_test) = split(TestFunction(Z))

  u_cn = cn(u_old, u_new)
  p_cn = cn(p_old, p_new)

  n = FacetNormal(mesh)

  while t < T:
    t += dt

    # Shallow water operator
    L = (inner(Dt(u_old, u_new, dt), u_test)*dx                    # u-time
      + inner(u_test, coriolis(u_cn))*dx                           # Coriolis
      + g*H*inner(u_test, grad(p_cn))*dx                           # pressure gradient
      + inner(avg(u_test), jump(p_cn, n))*dS                       #
      + inner(Dt(p_old, p_new, dt), p_test)*dx                     # p-time
      - inner(u_cn, grad(p_test))*dx                               # velocity divergence
      + inner(avg(u_cn), jump(p_test, n))*dS)

    # Write it as a Newton solve, in case we want to do nonlinear later:
    F = replace(L, {z_new: z})
    J = derivative(F, z)
    solve(F == 0, z, J=J, annotate=annotate)

    z_old.assign(z, annotate=annotate)

    if annotate:
      store(z_old, t=t)

    adj_inc_timestep()

  return z_old

def get_vars(name, i, mass):
  fwd_var = libadjoint.Variable(name=name, timestep=i, iteration=0)
  adj_var = fwd_var.to_adjoint(J)
  fwd = adjointer.get_variable_value(fwd_var).data
  adj = adjointer.get_variable_value(adj_var).data

  if not mass:
    return (fwd, adj)
  else:
    # and we need to convert the adjoint into the right space.
    z_test = TestFunction(Z)
    z_trial = TrialFunction(Z)

    a = assemble(inner(z_test, z_trial)*dx)
    L = adj.vector()
    adj_trial = Function(Z)
    solve(a, adj_trial.vector(), L, annotate=False)
    return (fwd, adj_trial)

def estimate_error():
  # higher-order function spaces
  Vh = VectorFunctionSpace(mesh, "DG", 2)
  Ph = FunctionSpace(mesh, "CG", 3)

  time_errors = []

  for i in range(adjointer.timestep_count-1):

    fwd_minus, adj_minus = get_vars("State", i, mass=True)
    fwd_plus, adj_plus = get_vars("State", i+1, mass=True)
    f = File("results/adju_plus.pvd")
    f << adj_plus.split(deepcopy=True)[0]
    f = File("results/adju_minus.pvd")
    f << adj_minus.split(deepcopy=True)[0]

    (u_minus, p_minus) = fwd_minus.split(deepcopy=True)
    (u_plus, p_plus)   = fwd_plus.split(deepcopy=True)
    u = Function(V, 0.5*u_minus.vector() + 0.5*u_plus.vector())
    p = Function(P, 0.5*p_minus.vector() + 0.5*p_plus.vector())

    uh = Function(Vh)
    uh.extrapolate(u)
    ph = Function(Ph)
    ph.extrapolate(p)
    u_err = uh - interpolate(u, Vh)
    p_err = ph - interpolate(p, Ph)

    (adju_minus, adjp_minus) = adj_minus.split(deepcopy=True)
    (adju_plus, adjp_plus)   = adj_plus.split(deepcopy=True)
    adju = Function(V, 0.5*adju_minus.vector() + 0.5*adju_plus.vector())
    adjp = Function(P, 0.5*adjp_minus.vector() + 0.5*adjp_plus.vector())

    adjuh = Function(Vh)
    adjuh.extrapolate(adju)
    adjph = Function(Ph)
    adjph.extrapolate(adjp)
    adju_err = adjuh - interpolate(adju, Vh)
    adjp_err = adjph - interpolate(adjp, Ph)

    # Now we have both the forward and adjoint variables here. Hooray!

    error = (dt * g * inner(grad(p_err), adju_err)*dx
          +  dt * inner(coriolis(u_err), adju_err)*dx
          -  dt * H * inner(grad(adjp_err), u_err)*dx)
#          +  dt * H * inner(div(u_err),  adjp_err)*dx)

    print "term 1: ", assemble(dt * g * inner(grad(p_err), adju_err)*dx)
    print "term 2: ", assemble(dt * inner(coriolis(u_err), adju_err)*dx)
    print "term 3: ", assemble(-dt * H * inner(grad(adjp_err), u_err)*dx)
    print "term 4: ", assemble(dt * H * inner(div(u_err),  adjp_err)*dx)
    time_errors.append(assemble(error))

  print time_errors
  return sum(time_errors)

if __name__ == "__main__": 
  ic = compute_ic()

  z = main(ic, annotate=True)

  adj_html("forward.html", "forward")
  adj_html("adjoint.html", "adjoint")

  u = split(z)[0]
  J = FinalFunctional(dot(Constant((1, 1)), u)*dx, name="VelocityNorm")
  print "Functional value: ", assemble(dot(Constant((1, 1)), u)*dx)

#  pvd = File("results/adjoint_us.pvd")
#  for (adj, adj_var) in compute_adjoint(J, forget=None):
#    pvd << adj.split(deepcopy=True)[0]
#    pass
#
#  dJdic = Function(adj)
#
##  def J(ic):
##    z = main(ic, annotate=False)
##    u = split(z)[0]
##    return assemble(dot(Constant((1, 1)), u)*dx)
##
##  minconv = test_initial_condition_adjoint(J, compute_ic(), dJdic, seed=10.0)
#
#  err = estimate_error()
#  print "Error estimate:   ", err
