from dolfin import *

R = 1500
S = 1500
mesh = Rectangle(0, -S, R, S, 50, 50)
V = VectorFunctionSpace(mesh, "CG", 2)
Q = FunctionSpace(mesh, "CG", 1)
Z = MixedFunctionSpace([V, Q])

free_top_bc = DirichletBC(Z.sub(0).sub(1), 0.0, "on_boundary && std::abs(x[1] - %s) < DOLFIN_EPS" % S)
free_bottom_bc = DirichletBC(Z.sub(0).sub(1), 0.0, "on_boundary && std::abs(x[1] + %s) < DOLFIN_EPS" % S)
free_left_bc = DirichletBC(Z.sub(0).sub(0), 0.0, "on_boundary && x[0] < DOLFIN_EPS")
free_right_bc = DirichletBC(Z.sub(0).sub(0), 0.0, "on_boundary && std::abs(x[0] - %s) < DOLFIN_EPS" % R)
free_bcs = [free_top_bc, free_bottom_bc, free_left_bc, free_right_bc]

noslip_bc = DirichletBC(Z.sub(0), (0.0, 0.0), "on_boundary")
noslip_bcs = [noslip_bc]

def store(z, name):

  u_pvd = File("results/%s_velocity.pvd" % name)
  p_pvd = File("results/%s_pressure.pvd" % name)
  (u, p) = z.split()

  u_pvd << (u)
  p_pvd << (p)

def coriolis(u, f):
  return f*as_vector([-u[1], u[0]])

def main(bcs, name):
  z = Function(Z)
  (u, p) = split(z)

  (u_test, p_test) = split(TestFunction(Z))

  f = Expression("f0 + beta*x[1]", f0=10**-4, beta=2*10**-4)
  nu = Constant(540)
  tau = Expression("-tau0 * cos(pi * x[1] / R)", tau0=0.1, R=R)

  L = (-div(u_test)*p*dx +
       inner(coriolis(u, f), u_test)*dx +
       nu*inner(grad(u), grad(u_test))*dx + 
       -inner(as_vector([tau, 0]), u_test)*dx +
       inner(p_test, div(u))*dx)

  J = derivative(L, z)
  solve(L == 0, z, bcs=bcs, J=J)
  store(z, name)
  return z

if __name__ == "__main__":
  z = main(free_bcs, name="free_slip")
  z = main(noslip_bcs, name="no_slip")
