from dolfin import *
from dolfin_adjoint import *

parameters["std_out_all_processes"] = False
parameters["form_compiler"]["quadrature_degree"] = 5
dolfin.parameters["adjoint"]["record_all"] = True
dolfin.parameters["adjoint"]["fussy_replay"] = True

mesh = Mesh("doughnut.xml.gz")
facet_labels = MeshFunction("uint", mesh, "doughnut_facets.xml.gz")

# define our function spaces: P2-P1 for velocity-pressure, P1 for advection
V = VectorFunctionSpace(mesh, "CG", 2)
P = FunctionSpace(mesh, "CG", 1)
Q = FunctionSpace(mesh, "CG", 1)
Z = MixedFunctionSpace([V, P, Q])

# Let's also gather the boundary conditions here
outer_u_bc = DirichletBC(Z.sub(0), (0, 0), facet_labels, 22) # no-slip for now
inner_u_bc = DirichletBC(Z.sub(0), (0, 0), facet_labels, 21) # no-slip
outer_T_bc = DirichletBC(Z.sub(2), 0, facet_labels, 22)
inner_T_bc = DirichletBC(Z.sub(2), 1, facet_labels, 21)
bcs = [outer_u_bc, inner_u_bc, outer_T_bc, inner_T_bc]

Ra = Constant(10**4)
g = Function(interpolate(Expression(("-x[0]/sqrt(x[0]*x[0] + x[1]*x[1])", "-x[1]/sqrt(x[0]*x[0] + x[1]*x[1])"), element=V.ufl_element()), V))
eta0 = Constant(1.0)
b = 0

dt = 0.0015
theta = 0.5

def Dt(z_old, z_new, dt):
  return (z_new - z_old)/Constant(dt)

def cn(z_old, z_new): # Crank-Nicolson
  return (1 - theta)*z_old + theta*z_new

def eta(T):
  if float(b) != 0:
    return eta0*exp(-b*T)
  else:
    return eta0

def strain(v):
  return 0.5*(grad(v) + grad(v).T)

if __name__ == "__main__":
  z_old = Function(Z, "newton_failure.xml.gz")
  (u_old, p_old, temp_old) = split(z_old)

  z_new = TrialFunction(Z)
  (u_new, p_new, temp_new) = split(z_new)

  z = Function(Z)
  z.assign(z_old)
  (u, p, temp) = split(z)

  (u_test, p_test, temp_test) = split(TestFunction(Z))

  u_cn = cn(u_old, u_new)
  p_cn = cn(p_old, p_new)
  temp_cn = cn(temp_old, temp_new)
  temp_nonlin = cn(temp_old, temp)
  u_nonlin = cn(u_old, u)

  # linearised Stokes-temperature operator, suitable for Picard iterations:
  L = (2.0 * eta(temp_nonlin) * inner(strain(u_cn), strain(u_test)) * dx
    + div(u_test)*p_cn*dx
    + div(u_cn)*p_test*dx 
    + Ra*temp_cn*inner(g, u_test)*dx
    + inner(Dt(temp_old, temp_new, dt), temp_test)*dx
    + inner(temp_test, inner(u_nonlin, grad(temp_cn)))*dx
    + inner(grad(temp_test), grad(temp_cn))*dx)

  # And then a Newton solve:
  F = replace(L, {z_new: z})
  J = derivative(F, z)
  solve(F == 0, z, bcs=bcs, J=J)
