from dolfin import *
from orkney_data import NetCDFExpression, TidalHarmonics, transform_mesh_latlon2xy, tidal_gauge_locations, write_detectors, get_bbox, open_detector_file
from numpy import maximum

set_log_level(PROGRESS)
mesh_file_name = "Orkney_coarse.xml"
facet_file_name = "Orkney_coarse_facet_region_new.xml"
#bathymetry_file_name = "/data/stephan/projects/amcg-tides/Orkney_Islands/mesh/G1.3/maps/wider_area_bathymetry.nc"
#tidal_file_name = "/data/stephan/projects/amcg-tides/Orkney_Islands/data/ap.ES2008.nc"
#gauge_file_name = '/data/stephan/projects/amcg-tides/Orkney_Islands/mesh/G1.3/maps/tidal_gauge_locations.dat'
bathymetry_file_name = "wider_area_bathymetry.nc"
tidal_file_name = "ap.ES2008.nc"
gauge_file_name = 'tidal_gauge_locations.dat'

minimum_depth = 7.0
nonlinear_iterations=2
have_advection=False
have_nonlinear_depth=False
have_rt0=False # if true switches on internal boundary integrals for pressure gradient, otherwise use P1DG-P2

# physical parameters
g = 9.81
c0 = 0.0025
f = 2*2*pi/86400*cos(59.*pi/180.) # coriolis at latitude 59 deg
# temporal settings
dt = 100
dump_period = 100
T  = 86400*30

mesh = Mesh(mesh_file_name)
latlon_mesh = Mesh(mesh)
surf_ids = MeshFunction("uint", mesh, facet_file_name)
ds = Measure("ds")[surf_ids]

if have_rt0:
  V = FunctionSpace(mesh, "Raviart-Thomas", 1)
  P = FunctionSpace(mesh, "DG", 0)
else:
  V = VectorFunctionSpace(mesh, "DG", 1)
  P = FunctionSpace(mesh, "CG", 2)
Z = MixedFunctionSpace([V, P])

# initialise depth as Function instead of as Expression, so we can massage it once
bath_expression = NetCDFExpression(bathymetry_file_name, "z")
X = FunctionSpace(mesh, "CG", 1)
bath = Function(X)
bath.assign(bath_expression)
depth = conditional( gt(-bath, minimum_depth), -bath, minimum_depth)

H_pvd = File("results/H.pvd")
H_pvd << (bath, 0.0)

x_new=TrialFunction(X)
x_test=TestFunction(X)
x=Function(X)
solve(x_test*x_new*dx==x_test*depth*dx, x)
H_pvd << (x, 1.0)

if have_advection:
  # little ramp function, so that advection is only applied away from the boundaries
  class InteriorExpression(Expression):
    def eval(self, values, X):
      d=min(abs(X[0]--4.45), abs(X[1]-57.35))
      if d>0.25:
        values[0]=1.0
      elif d>0.05:
        values[0]=(d-0.05)/0.2
      else:
        values[0]=0.0

  interior = Function(X)
  interior.assign(InteriorExpression())
  i_pvd = File("results/interior.pvd")
  i_pvd << (interior, 0.0)

# initialise TidalHarmonics instance that stores the complex tidal components as Function on P
# (actually only need it on part of the boundary but can't figure out how to do this)
tidal_harmonics = TidalHarmonics(P, tidal_file_name)

# transform the mesh *after* initialising the bathymetry and NetCDF fields that are function of lat and lon (bit of a hack but fine for now)
bbox = get_bbox(mesh)
transform_mesh_latlon2xy(mesh)

tidal_gauge_xy = tidal_gauge_locations(gauge_file_name, bbox)
detector_file = open_detector_file('results/orkney.detectors')
u_pvd = File("results/u.pvd")
p_pvd = File("results/p.pvd")

def store(z, t):
  t=float(t)
  print "Storing variables at t=%s" % t
  (u, p) = z.split(deepcopy=True)
  u_pvd << (u, t)
  p_pvd << (p, t)

def cn(old, new):
  return 0.0*old + 1.0*new

def coriolis(u, f):
  return f*as_vector([-u[1], u[0]])

def main():

  z_new = TrialFunction(Z)
  (u_new, p_new) = split(z_new)

  z_old = Function(Z)
  (u_old, p_old) = split(z_old)

  z_test = TestFunction(Z)
  (u_test, p_test) = split(z_test)

  z = Function(Z)
  (u, p) = split(z)

  u_cn = cn(u_old, u_new)
  p_cn = cn(p_old, p_new)
  u_nonlin = cn(u_old, u)
  p_nonlin = cn(p_old, p)

  t = 0

  if have_nonlinear_depth:
    H = depth + p_nonlin
  else:
    H = depth
  u_norm = sqrt(inner(u_nonlin, u_nonlin))

  n = FacetNormal(mesh)
  tidal_fs = Function(P)

  F = (inner((u_new - u_old)/dt, u_test)*dx
    + g*inner(grad(p_cn), u_test)*dx
    - (p_cn-tidal_fs)*g*inner(u_test, n)*ds(1009)
    - (p_cn-tidal_fs)*g*inner(u_test, n)*ds(1010)
    - (p_cn-tidal_fs)*g*inner(u_test, n)*ds(1011)
    - (p_cn-tidal_fs)*g*inner(u_test, n)*ds(1013)
    + c0*u_norm*(1.0/H)*inner(u_cn, u_test)*dx
    + inner(coriolis(u_cn, f), u_test)*dx
    + inner((p_new - p_old)/dt, p_test)*dx
    - inner(grad(p_test), H*u_cn)*dx
    + p_test*H*inner(u_cn, n)*ds(1009)
    + p_test*H*inner(u_cn, n)*ds(1010)
    + p_test*H*inner(u_cn, n)*ds(1011)
    + p_test*H*inner(u_cn, n)*ds(1013)
    )

  #F = replace(F, {z_new: z})
  if have_advection:
    u_adv = u_nonlin*interior
    # if(u.n<0, u.n, 0.0)
    udown = (dot(u_adv,n)-abs(dot(u_adv,n)))/2.0
    F += (dot(u_test, dot(u_nonlin, nabla_grad(u_cn)))*dx
        - dot(u_test('+'),udown('+')*jump(u_cn))*dS
        + dot(u_test('-'),udown('-')*jump(u_cn))*dS
        - dot(u_test,udown*u_cn)*ds(1014))

  if have_rt0:
    F += g*inner(jump(p_cn, n), avg(u_test))*dS
    - inner(jump(p_test, n), avg(H)*avg(u_cn))*dS

  Flhs = lhs(F)
  Frhs = rhs(F)

  store(z,t)
  write_detectors(detector_file, tidal_gauge_xy, z, t)

  while t < T:
    tidal_harmonics.get_tidal_values(tidal_fs, t+0.5*dt)

    for it in range(nonlinear_iterations):

      solve(Flhs == Frhs, z, 
          solver_parameters={"linear_solver": "mumps"})

    z_old.assign(z)
    t = t + dt
    
    if abs(t%dump_period)<1e-5*dt:
      store(z,t)
    write_detectors(detector_file, tidal_gauge_xy, z, t)

  return z_old

if __name__ == "__main__":
  z = main()
