from dolfin import *
from numpy import isnan,cos,sin,loadtxt,array,nan

tidal_frequencies={'q1':  0.64959e-04, 'o1':  0.67598e-04, 'p1': 0.72523e-04, 'k1': 0.72921e-04, 'n2':  1.3788e-04, 
    'm2': 1.40519e-04, 's2':  1.45444e-04, 'k2': 1.45842e-04}
Rearth=5**8*6*exp(1) # xkcd approved

class NetCDFReader(object):
  def __init__(self, filename):
    from Scientific.IO.NetCDF import NetCDFFile
    self.nc = NetCDFFile(filename)
    if 'longitude' in self.nc.variables:
      lon = self.nc.variables['longitude'][:]
    else:
      lon = self.nc.variables['lon'][:]
    self.lon_min = lon[0]
    self.delta_lon = len(lon)/(lon[-1]-lon[0])
    if 'latitude' in self.nc.variables:
      lat = self.nc.variables['latitude'][:]
    else:
      lat = self.nc.variables['lat'][:]
    self.lat_min = lat[0]
    self.delta_lat = len(lat)/(lat[-1]-lat[0])

  def set_field(self,field):
    self.val = self.nc.variables[field]

  def get_val(self, lat, lon):
    i = int(round((lat-self.lat_min)*self.delta_lat))
    alpha = (lat-self.lat_min)*self.delta_lat-float(i)
    j = int(round((lon-self.lon_min)*self.delta_lon))
    beta = (lon-self.lon_min)*self.delta_lon-float(j)
    try:
      value = ((1.0-beta)*((1.0-alpha)*self.val[i,j]+alpha*self.val[i+1,j])+
        beta*((1.0-alpha)*self.val[i,j+1]+alpha*self.val[i+1,j+1]))
    except:
      print lat, lon, i, j
      raise
    if isnan(value):
      return 0.0
    else:
      return value

class NetCDFReaderExpression(Expression):
  """Dolfin Expression initialised from a NetCDFReader object.
  Needs evaluating on a mesh with lat,lon coordinates."""
  def __init__(self,nc):
    self.nc = nc

  def eval(self, values, x):
    values[0] = self.nc.get_val(x[1], x[0])

class NetCDFExpression(Expression):
  """Dolfin Expression initialised from field in a NetCDF file.
  Needs evaluating on a mesh with lat,lon coordinates."""
  def __init__(self, filename, field):
    self.nc = NetCDFReader(filename)
    self.nc.set_field(field)

  def eval(self, values, x):
    values[0] = self.nc.get_val(x[1], x[0])

class TidalHarmonics(object):
  def __init__(self, function_space, filename):
    nc=NetCDFReader(filename)
    self.real_components={}
    self.imag_components={}
    for comp in tidal_frequencies.keys():
      nc.set_field(comp+'amp')
      amp = nc.val[:]
      nc.set_field(comp+'phase')
      phase= nc.val[:]*pi/180.
      # real component
      nc.val = amp*cos(phase)
      ncexp = NetCDFReaderExpression(nc)
      self.real_components[comp] = Function(function_space)
      self.real_components[comp].assign(ncexp)
      # imaginary component
      nc.val = -amp*sin(phase)
      ncexp = NetCDFReaderExpression(nc)
      self.imag_components[comp] = Function(function_space)
      self.imag_components[comp].assign(ncexp)

  def get_tidal_values(self, f, t):
    fvec=f.vector()
    fvec.zero()
    for comp,omega in tidal_frequencies.items():
      fvec.axpy(cos(omega*t), self.real_components[comp].vector())
      fvec.axpy(sin(omega*t), self.imag_components[comp].vector())

def get_bbox(mesh):
  lon = mesh.coordinates()[:,0]
  lat = mesh.coordinates()[:,1]
  return MPI.min(lon.min()), MPI.max(lon.max()), MPI.min(lat.min()), MPI.max(lat.max())

def transform_mesh_latlon2xy(mesh):
  lon = mesh.coordinates()[:,0]
  lat = mesh.coordinates()[:,1]
  rescaling = pi/180. * Rearth
  mesh.coordinates()[:,0] = lon*rescaling*cos(lat*pi/180.)
  mesh.coordinates()[:,1] = lat*rescaling

def tidal_gauge_locations(filename, bbox):
  data = loadtxt(filename)
  tidal_gauge_xy = data[:,:2]

  lonmin, lonmax, latmin, latmax = bbox
  mask=[]
  for x,y in tidal_gauge_xy:
    mask.append(lonmin<x<lonmax and latmin<y<latmax)
  mask=array(mask)
  print "Selected %(ngauges)d in boundingbox %(bbox)s" % {'ngauges':sum(mask), 'bbox':str(bbox)}
  tidal_gauge_xy=tidal_gauge_xy[mask,:]
  
  # apply same rescaling as in transform_mesh_latlon2xy
  rescaling = pi/180. * Rearth
  tidal_gauge_xy[:,0] = tidal_gauge_xy[:,0]*rescaling*cos(tidal_gauge_xy[:,1]*pi/180.)
  tidal_gauge_xy[:,1] = tidal_gauge_xy[:,1]*rescaling
  return tidal_gauge_xy

def open_detector_file(detector_file_name):
  if MPI.process_number()==0:
    detector_file = file(detector_file_name, 'w')
  else:
    # need to pass something into write_detectors
    detector_file = None
  return detector_file

def write_detectors(detector_file, tidal_gauge_xy, z, t):
  print "Writing detectors at t=%s" % t
  line = '%f '%t
  rank = MPI.process_number()
  nprocs = MPI.num_processes()
  for xy in tidal_gauge_xy:
    if MPI.num_processes>1:
      try:
        values = z(xy)
        # found by this process
        found = rank
      except RuntimeError:
        # not found by this process
        found = nprocs
      # lowest rank process that sucessfully probed this location
      min_found = MPI.min(found)
      if min_found==rank:
        # found by me, nothing to do
        pass
      elif min_found<nprocs:
        # found by someone else
        values = array([0.,0.,0.])
      else:
        # found by noone
        values = array([nan,nan,nan])
      values=array([MPI.sum(val) for val in values])
    else:
      # serial case
      try:
        values = z(xy)
      except RuntimeError:
        values = array([nan,nan,nan])
    line += '%f %f %f '%tuple(values)
  if rank==0:
    detector_file.write(line[:-1]+'\n')
    detector_file.flush()
