import matplotlib
matplotlib.use('SVG')

from dolfin import *
import sys
import shutil

parameters["std_out_all_processes"] = False
parameters["form_compiler"]["quadrature_degree"] = 5

mesh = Mesh("doughnut.xml.gz")
facet_labels = MeshFunction("uint", mesh, "doughnut_facets.xml.gz")

# define our function spaces: P2-P1 for velocity-pressure, P1 for advection
V = VectorFunctionSpace(mesh, "CG", 2)
P = FunctionSpace(mesh, "CG", 1)
Q = FunctionSpace(mesh, "CG", 1)
Z = MixedFunctionSpace([V, P, Q])

if MPI.process_number() == 0:
  print "Problem size: %s = (%s, %s, %s)" % (Z.dim(), V.dim(), P.dim(), Q.dim())

# Let's also gather the boundary conditions here
outer_u_bc = DirichletBC(Z.sub(0), (0, 0), facet_labels, 22) # no-slip for now
inner_u_bc = DirichletBC(Z.sub(0), (0, 0), facet_labels, 21) # no-slip
outer_T_bc = DirichletBC(Z.sub(2), 0, facet_labels, 22)
inner_T_bc = DirichletBC(Z.sub(2), 1, facet_labels, 21)
noslip_bcs = [outer_u_bc, inner_u_bc, outer_T_bc, inner_T_bc]

Ra = Constant(10**3)
g = Function(interpolate(Expression(("-x[0]/sqrt(x[0]*x[0] + x[1]*x[1])", "-x[1]/sqrt(x[0]*x[0] + x[1]*x[1])"), element=V.ufl_element()), V))
eta0 = Constant(1.0)
b = Constant(ln(1000))
#b = 0.0
target_cfl = 3.0

ref_dt = 0.0015
T = 1.0

theta = 0.5

temp_pvd = File("results/temperature.pvd")
u_pvd = File("results/velocity.pvd")
p_pvd = File("results/pressure.pvd")
g_pvd = File("results/gravity.pvd")
g_pvd << (g, 0.0)

if MPI.process_number() == 0:
  shutil.copy(sys.argv[0], "results/")

def store(z, t, i):
  t = float(t)
  if MPI.process_number() == 0:
    print "Storing variables at t=%s" % t

  (u, p, temp) = z.split()

  temp_pvd << (temp, t)
  u_pvd << (u, t)
  p_pvd << (p, t)

  f = File("results/state%06d.xml.gz" % i)
  f << z

def print_cfl(z, mesh, dt):
  import numpy
  u = z.split(deepcopy=True)[0]

  umax = numpy.max(numpy.abs(u.vector().array()))
  hmin = mesh.hmin()
  CFL = MPI.max(umax * dt/hmin)

  if MPI.process_number() == 0:
    if CFL < target_cfl:
      info_blue("CFL number: %s" % CFL)
    else:
      info_red("CFL number: %s" % CFL)

  if target_cfl is not None:
    new_dt = MPI.min(min(target_cfl*hmin/umax, 0.1))
    if MPI.process_number() == 0:
      info_blue("New dt: %s" % new_dt)
    return new_dt
  else:
    return dt

def Dt(z_old, z_new, dt):
  return (z_new - z_old)/Constant(dt)

def cn(z_old, z_new): # Crank-Nicolson
  return (1 - theta)*z_old + theta*z_new

def eta(T):
  if float(b) != 0:
    return eta0*exp(-b*T)
  else:
    return eta0

def strain(v):
  return 0.5*(grad(v) + grad(v).T)

def save_dts(times, dts):
  if MPI.process_number() == 0:
    data = """
times = %s
dts = %s
  """ % (repr(times), repr(dts))
    f = open("results/timesteps.log", "w")
    f.write(data)
    f.close()

    try:
      import pylab
      pylab.plot(times, dts)
      pylab.savefig('results/timesteps.svg')
    except:
      pass

def initial_guess(z_old, bcs, dt):
  # Do 5 Picard iterations to get a reasonable guess
  (u_old, p_old, temp_old) = split(z_old)

  z_new = TrialFunction(Z)
  (u_new, p_new, temp_new) = split(z_new)

  z = Function(Z)
  z_tmp = Function(Z)
  (u, p, temp) = split(z)

  (u_test, p_test, temp_test) = split(TestFunction(Z))

  temp_cn = cn(temp_old, temp_new)
  u_nonlin = cn(u_old, u)

  for i in range(5):
    L = (2.0 * eta(temp) * inner(strain(u_new), strain(u_test)) * dx
      + div(u_test)*p_new*dx
      + div(u_new)*p_test*dx 
      + Ra*temp_new*inner(g, u_test)*dx
      + inner(Dt(temp_old, temp_new, dt), temp_test)*dx
      + inner(temp_test, inner(u_nonlin, grad(temp_cn)))*dx
      + inner(grad(temp_test), grad(temp_cn))*dx)

    solve(lhs(L) == rhs(L), z_tmp, bcs)
    z.assign(z_tmp)

  return z
def main(ic, t, bcs):
  dt = ref_dt
  i = 0

  times = [t]
  dts   = [dt]

  store(ic, t=t, i=i)

  z_old = ic
  (u_old, p_old, temp_old) = split(z_old)

  z_new = TrialFunction(Z)
  (u_new, p_new, temp_new) = split(z_new)

  z = initial_guess(z_old, bcs, dt)
  (u, p, temp) = split(z)

  (u_test, p_test, temp_test) = split(TestFunction(Z))

  temp_cn = cn(temp_old, temp_new)
  u_nonlin = cn(u_old, u)

  while t < T:
    t += dt

    # linearised Stokes-temperature operator, suitable for Picard iterations:
    L = (2.0 * eta(temp) * inner(strain(u_new), strain(u_test)) * dx
      + div(u_test)*p_new*dx
      + div(u_new)*p_test*dx 
      + Ra*temp_new*inner(g, u_test)*dx
      + inner(Dt(temp_old, temp_new, dt), temp_test)*dx
      + inner(temp_test, inner(u_nonlin, grad(temp_cn)))*dx
      + inner(grad(temp_test), grad(temp_cn))*dx)

    # And then a Newton solve:
    F = replace(L, {z_new: z})
    J = derivative(F, z)

    try:
      solve(F == 0, z, bcs=bcs, J=J, solver_parameters={"linear_solver": "mumps"})
    except RuntimeError:
      save_dts(times, dts)

      solvers = ["superlu_dist", "spooles"]
      for solver in solvers:
        info_red("Solver failed, trying %s" % solver)
        z.assign(z_old)
        try:
          solve(F == 0, z, bcs=bcs, J=J, solver_parameters={"linear_solver": solver})
          break
        except RuntimeError:
          if solver == solvers[-1]:
            raise

    z_old.assign(z)
    dt = print_cfl(z_old, mesh, dt)

    i = i + 1
    store(z_old, t=t, i=i)

    times.append(t)
    dts.append(dt)

  save_dts(times, dts)
  return z_old

def get_ic(bcs):
#  class StateIC(Expression):
#    def eval(self, values, X):
#      values[0] = 0
#      values[1] = 0
#      values[2] = 0
#      values[3] = 0
#
#      if (X[0] < -1.22 or X[0] > 1.22):
#        if (X[1] < 1.22 and X[1] > -1.22):
#          values[3] = 1.
#      if (X[1] < -1.22 or X[1] > 1.22):
#        if (X[0] < 1.22 and X[0] > -1.22):
#          values[3] = 1.
#
#    def value_shape(self):
#      return (4,)
#
#  ic = Function(Z)
#  ic.interpolate(StateIC())
#  [bc.apply(ic.vector()) for bc in bcs]
  
  ic_temp = split(Function(Z, "state000333.xml.gz"))[2]

  (u_ic, p_ic, t_ic) = split(TrialFunction(Z))
  (u_test, p_test, t_test) = split(TestFunction(Z))

  a = inner(u_test, u_ic)*dx + inner(p_test, p_ic)*dx + inner(t_test, t_ic)*dx
  L = inner(Constant((0,0)), u_test)*dx + inner(Constant(0), p_test)*dx + inner(t_test, ic_temp)*dx

  z = Function(Z)
  solve(a == L, z)
  
  return z

if __name__ == "__main__": 
  ic = get_ic(noslip_bcs)
  ic_copy = Function(ic)

  z = main(ic, t=0.0, bcs=noslip_bcs)

#  adj_html("forward.html", "forward")
#  adj_html("adjoint.html", "adjoint")
#
#  u = split(z)[0]
#  J = FinalFunctional(inner(u, u)*dx)
#
#  dJdic = compute_gradient(J, InitialConditionParameter("State"))
#
#  splits = dJdic.split(deepcopy=True)
#
#  grad_u = File("results/grad_velocity.pvd")
#  grad_u << splits[0]
#  grad_p = File("results/grad_pressure.pvd")
#  grad_p << splits[1]
#  grad_T = File("results/grad_temperature.pvd")
#  grad_T << splits[2]

#  def J(ic):
#    z = main(ic)
#    u = split(z)[0]
#    return assemble(inner(u, u)*dx)
#
#  pdir = Function(Z)
#  pdir.vector()[:] = 1.0
#
#  minconv = test_initial_condition_adjoint(J, ic_copy, dJdic, seed=1.0e-3, perturbation_direction=pdir)

